/*** Copyright: OGIHARA Takeshi, Nov. 2012 ***/

#import "Solver.h"
#import "math.h"

@implementation Solver

- (void)solveA:(double)a B:(double)b C:(double)c
{
    NSString *s = @"";
    if (a == 0.0) {
	if (b == 0.0)
		s = @"No roots";
	else
		s = [NSString stringWithFormat: @"root = %.4f", -c / b];
    }else {
	double s1, s2;
	double d = b * b - 4.0 * a * c;
	if (d > 0.0) {
		d = sqrt(d);
		s1 = (-b + d) / (2.0 * a);
		s2 = (-b - d) / (2.0 * a);
		s = [NSString stringWithFormat:
			@"roots = %.4f, %.4f", s1, s2];
	}else {
		double w = -b / (2.0 * a);
		if (d == 0.0)
			s = [NSString stringWithFormat: @"root = %.4f", w];
		else /* d < 0.0 */ {
			d = sqrt(-d) / (2.0 * a);
			s = [NSString stringWithFormat:
				@"roots = %.4f + %.4fi, %.4f - %.4fi",
				w, d, w, d];
		}
	}
    }
    [text setStringValue: s];
}

static int coefficientsChanged(coeffs_coval_t *cp, Solver *me)
{
    double a = CovalMemberValue(cp, a);
    double b = CovalMemberValue(cp, b);
    double c = CovalMemberValue(cp, c);
    [me solveA:a B:b C:c];
    return COVAL_PROC_SUCCESS;
}


- (id) init {
	if ((self = [super init]) != nil) {
		CovalInit(&coef);
		CovalSetCallback(&coef, coefficientsChanged, self);
	}
	return self;
}

- (coeffs_coval_t *)covalOfCoefficients {
	return &coef;
}

- (void)setTextfield:(NSTextField *)obj {
	text = obj;
}

@end
