//
//  coval.h
//
//  coval version 0.8
//      2012.11.01 Takeshi Ogihara
//  Copyright (C) 2010, 2011, 2012 Takeshi Ogihara
//
#ifndef __COVAL_H__
#define __COVAL_H__

#if !defined(NULL)
#define  NULL   (void *)0
#endif

// You can define USE_COVAL_BRIDGE as 0 not to use covalbridge.
// With option -DUSE_COVAL_BRIDGE=0, only coval routines are compiled.
#ifndef USE_COVAL_BRIDGE
#define USE_COVAL_BRIDGE 1
#endif

#define COVAL_PROC_SUCCESS  0   // callback is done successfully
#define COVAL_PROC_IGNORED  1   // callback does no practical work
#define COVAL_PROC_ERROR    (-1)  // failure

struct _coval_proto;
typedef int (*coval_callback)(struct _coval_proto *, void *);
typedef int (*coval_update)(struct _coval_proto *, void *, void *);

typedef struct _coval_center_info {
    unsigned short _count;     // reference counter of center coval
    unsigned char  _properties;
    unsigned char  _inLoop;    // loop is in process
} coval_center_info;

typedef struct _coval_proto {
    void *_p;  // if NULL, not bound
    union {
        struct _coval_proto *_cent; // link from member to center
        coval_center_info _cinfo;   // info. of center coval
    } _b;
    unsigned char _ringflag;
    coval_callback _cb; // callback
    void *_ag; // argument for callback func
    coval_update _up; // update
    void *_au; // argument for update func
    struct _coval_proto *_nx; // link to next element
    struct _coval_proto *_loopcntl;
        // center->_loopcntl is the last element of the ring.
        // center->_nx->_loopcntl is current pointer of calling loop.
    struct _coval_proto *_upptr;  // element with update function
    long _d; // dummy for data
} coval_base;
typedef coval_base *coval_ref;

#define CovalInit(Cp)   (Cp)->_p = NULL, (Cp)->_cb = 0, (Cp)->_up = 0

#define CovalTypeOf(TYP)    struct { TYP *_p; \
    union { coval_base *_cent; coval_center_info _cinfo; } _b; \
    unsigned char _ringflag; \
    coval_callback _cb; void *_ag; coval_update _up; void *_au; \
    coval_base *_nx, *_loopcntl, *_upptr; \
    TYP _d; }

void _coval_call(coval_base *, int);
int _coval_request(coval_base *);
#define CovalValue(Cp)  ( (Cp)->_p ? *((Cp)->_p) : (Cp)->_d )
#define CovalGetValue(Cp)  (_coval_request((coval_base *)(Cp)) ? *((Cp)->_p) : (Cp)->_d )
#define CovalMemberValue(Cp, M) ((Cp)->_p ? ((Cp)->_p)->M : (Cp)->_d.M)
#define CovalGetMemberValue(Cp, M)  \
    (_coval_request((coval_base *)(Cp)) ? ((Cp)->_p)->M : (Cp)->_d.M )
#define CovalAssign(Cp, V)  if ( (Cp)->_p ) { \
    *((Cp)->_p) = (V); _coval_call((coval_base *)(Cp), 0); \
    }else (Cp)->_d = (V)
#define CovalSetValue(Cp, V)  if ( (Cp)->_p ) { \
    *((Cp)->_p) = (V); _coval_call((coval_base *)(Cp), 1); \
    }else do{ (Cp)->_d = (V); if ((Cp)->_cb) (Cp)->_cb((coval_base *)(Cp), (Cp)->_ag); \
    }while(0)
#define CovalAssignMember(Cp, M, V)   if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); _coval_call((coval_base *)(Cp), 0); \
    }else (Cp)->_d.M = (V)
#define CovalAssignMembers2(Cp, M, V, M2, V2)   if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); \
    _coval_call((coval_base *)(Cp), 0); \
    }else (Cp)->_d.M = (V), (Cp)->_d.M2 = (V2)
#define CovalAssignMembers3(Cp, M, V, M2, V2, M3, V3)   if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); \
    _coval_call((coval_base *)(Cp), 0); \
    }else (Cp)->_d.M = (V), (Cp)->_d.M2 = (V2), (Cp)->_d.M3 = (V3)
#define CovalAssignMembers4(Cp, M, V, M2, V2, M3, V3, M4, V4)  \
    if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); \
    (Cp)->_p->M4 = (V4); _coval_call((coval_base *)(Cp), 0); \
    }else (Cp)->_d.M = (V), (Cp)->_d.M2 = (V2), \
        (Cp)->_d.M3 = (V3), (Cp)->_d.M4 = (V4)
#define CovalSetMemberValue(Cp, M, V)  if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); _coval_call((coval_base *)(Cp), 1); \
    }else do{ (Cp)->_d.M = (V); if ((Cp)->_cb) (Cp)->_cb((coval_base *)(Cp), (Cp)->_ag); \
    }while(0)
#define CovalSetMemberValues2(Cp, M, V, M2, V2)  if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); _coval_call((coval_base *)(Cp), 1); \
    }else do{ (Cp)->_d.M = (V); (Cp)->_d.M2 = (V2); \
        if ((Cp)->_cb) (Cp)->_cb((coval_base *)(Cp), (Cp)->_ag); }while(0)
#define CovalSetMemberValues3(Cp, M, V, M2, V2, M3, V3)  if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); \
    _coval_call((coval_base *)(Cp), 1); \
    }else do{ (Cp)->_d.M = (V); (Cp)->_d.M2 = (V2); (Cp)->_d.M3 = (V3); \
        if ((Cp)->_cb) (Cp)->_cb((coval_base *)(Cp), (Cp)->_ag); }while(0)
#define CovalSetMemberValues4(Cp, M, V, M2, V2, M3, V3, M4, V4)  if ( (Cp)->_p ) { \
    (Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); (Cp)->_p->M4 = (V4); \
    _coval_call((coval_base *)(Cp), 1); \
    }else do{ (Cp)->_d.M = (V); (Cp)->_d.M2 = (V2); \
              (Cp)->_d.M3 = (V3); (Cp)->_d.M4 = (V4); \
        if ((Cp)->_cb) (Cp)->_cb((coval_base *)(Cp), (Cp)->_ag); }while(0)

#define CovalSetCallback(Cp, F, A)  \
    (Cp)->_cb = (coval_callback)(F), (Cp)->_ag = (A)
#define SetCovalCallback    CovalSetCallback    // for compatibility (0.5)

#define CovalSetUpdate(Cp, F, A)  \
    (Cp)->_up = (coval_update)(F), (Cp)->_au = (A)

/* #define CovalSetCallSelf(Cp, F)  (Cp)->_callself = (F) */
#define SetCovalCallSelf    CovalSetCallSelf    // for compatibility (0.5)

void _bindCoval(coval_base *target, coval_base *cov, void *pp);
#define CovalBind(TARGp, Cp)    _bindCoval((coval_base *)(TARGp), (coval_base *)(Cp), NULL)
#define BindCoval   CovalBind   // for compatibility (0.5)

void _unbindCoval(coval_base *, unsigned int);
#define CovalUnbind(CV)   _unbindCoval((coval_base *)(CV), sizeof(*(CV)));
#define UnbindCoval     CovalUnbind     // for compatibility (0.5)

#define BIND_FLAG_BREAKABLE 0x01    // quit if COVAL_PROC_IGNORED
void _setBindingProperty(coval_base *, unsigned int, int);
#define CovalSetBindingLoopBreakable(Cp, F) \
    _setBindingProperty((coval_base *)(Cp), BIND_FLAG_BREAKABLE, (F))
#define SetBindingLoopBreakable  CovalSetBindingLoopBreakable
    // for compatibility (0.5)

#endif /* __COVAL_H__ */
